/*********************************************************************
 *
 *         				XSoko
 *
 *             Copyright (C)  1998  Wojciech Rawdanik
 *  
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
*	april	1998
 **********************************************************************/

#include <stdio.h>
#include <time.h> 
#include <sys/time.h> 
#include <unistd.h>
#include "EZ.h"

#define TILES_X 		20
#define TILES_Y 		20

#define TILE_SIZE 	24

#define FRAME_SIZE  4
#define STATUS_WIDTH 152

#define MAIN_WIDTH	(TILES_X * TILE_SIZE)+ (2 *FRAME_SIZE) + STATUS_WIDTH
#define MAIN_HEIGHT (TILES_Y * TILE_SIZE)+ (2 *FRAME_SIZE)

#define SOKOERROR_OPEN_FILE 		1
#define SOKOERROR_READ_FILE 		2
#define SOKOERROR_BAD_FORMAT  	3
#define SOKO_NO_SAVED_GAME			4

#define SOKO_LEFT 	0
#define SOKO_RIGHT	1
#define SOKO_UP 		2
#define SOKO_DOWN 	3

#define SOKO_PLAYBACKGROUND 	"lin64.gif"
#define	SOKO_MAPS	"levels/"
#define	SOKO_GRAPH	"graphics/"

//  Basic tile

typedef struct 
{
	char	typef;
	char	typeb;
	EZ_Widget *foreground;
	EZ_Widget *background;
} 	tile;

int		LoadLevelData(char *filename);
void	InitializePlayfield();
int 	CreatePlayfielsWidgets();
void  	DeletePlayfieldWidgets();
void  	SokoExit();
static	void MainEventHandler(EZ_Widget *widget, void *data, int etype, XEvent *xeve);
static	void GameEventHandler(EZ_Widget *widget, void *data, int etype, XEvent *xeve);
int 	ValidateHeroPosition(int posX,int posY,int direction);
int 	ValidateBlockPosition(int posX,int posY);
void	RedrawBlockPosition(int oldx, int oldy, int newx, int newy);
int		IsLevelDone();

void	CallBackStartGame(EZ_Widget *widget, void *data);
void  	CallBackRestartLevel(EZ_Widget *widget, void *data);
void	CallBackQuit(EZ_Widget *widget, void *data);
void	CallBackAbout(EZ_Widget *widget, void *data);
void	CallBackHelp(EZ_Widget *widget, void *data); 
int	  	DisplayDialog(char *info, char *buttonA, char *buttonB, char *buttonC,char *title);
void	DialogCallbackA(EZ_Widget *widget, int *answer);
void	DialogCallbackB(EZ_Widget *widget, int *answer);
void	DialogCallbackC(EZ_Widget *widget, int	*answer);
void	HelpCallbackCloseWindow(EZ_Widget *widget, void *data);
void  	CallBackSaveQuit(EZ_Widget *widget, void *data);
void	CallBackUndoMove(EZ_Widget *widget, void *data);

void	DisplayPlayground(int mode);

void	NextLevel();
void	InitializeGame();
void	StopGame();
void	StartGame();
void	RestartLevel();
int 	CreatePixmaps();
void	DisplayHelpWindow();
int 	SaveCurrentGame();
int 	LoadGame();

// Playfield map

tile	PTable[TILES_X][TILES_Y];


EZ_Widget *MainFrame, *LeftFrame, *RightFrame, *SubFrame;
EZ_Widget *AboutLabel, *AboutAutor;
EZ_Widget *BUndoMove, *BStartGame, *BStartLevel;
EZ_Widget *BHelp;
EZ_Widget *BQuit, *BSaveQuit;
EZ_Widget *Player, *Score, *Level;
EZ_Widget *Separator;

EZ_Widget *Playfield;


EZ_Pixmap 	*Wall,*Object,*Platform,*Logo,*Background;
EZ_Pixmap 	*Hero[4];

Atom       WMProtocolsAtom;  
Atom       DeleteWindowAtom;

int HeroPosX,HeroPosY;
int LOldBX,LOldBY,LNewBX,LNewBY;
int LOldHX,LOldHY,LNewHX,LNewHY;

int GameOn;
int	Barrels;

char	PlayerInfo[128];
char	PlayerHome[256];
char	LevelInfo[128];
int 	GameLevel;
long	Scorew;
int 	Moves;
char	path[512];

main	(int ac, char **av)
{
	
	GameOn=0;
	GameLevel=0;
	strcpy(PlayerInfo,"Player: ");
	strcat(PlayerInfo, getenv("LOGNAME") );
	strcat(PlayerHome, getenv("HOME"));
	
// Initialize display
	
	EZ_Initialize(ac,av,0);
	
	//EZ_DisableLabelOffset();
	EZ_DisableSliderDepression();
	EZ_SetDefaultLabelFont("-Adobe-Helvetica-Medium-R-Normal--*-120-*-*-*-*-*-*");
	EZ_SetDefaultEntryTextFont("-Adobe-Helvetica-Bold-R-Normal--*-120-*-*-*-*-*-*");
	
	if(fopen("/usr/local/games/XSoko/graphics/pingwin.xpm","r")==NULL)
		getcwd(path,511);
	else
		strcpy(path,"/usr/local/games/XSoko");
	
	if(CreatePixmaps())
	{
		DisplayDialog("Unable to load pixmaps.","OK" , NULL, NULL,"XSoko error");
	}	
	DeleteWindowAtom = EZ_GetAtom("WM_DELETE_WINDOW");
	WMProtocolsAtom = EZ_GetAtom("WM_PROTOCOLS");
	
	//* Create main widgets  
	
	MainFrame= EZ_CreateWidget(EZ_WIDGET_FRAME, NULL,
		EZ_WIDTH, MAIN_WIDTH,
		EZ_HEIGHT, MAIN_HEIGHT,
		EZ_PADX, 0,
		EZ_PADY, 0,
		EZ_WMHINTS,PMinSize|PMaxSize|PWinGravity,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		0); 
	
	LeftFrame= EZ_CreateWidget(EZ_WIDGET_FRAME, MainFrame,
		EZ_X, 0,
		EZ_Y, 0,	
		EZ_WIDTH, (TILES_X * TILE_SIZE) + (FRAME_SIZE * 2),
		EZ_HEIGHT, (TILES_Y * TILE_SIZE) + (FRAME_SIZE * 2),
		EZ_BORDER_WIDTH, FRAME_SIZE,
		EZ_BORDER_TYPE, EZ_BORDER_RIDGE,
		EZ_PADX, 0,
		EZ_PADY, 0,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		0);	
	
	
	RightFrame= EZ_CreateWidget(EZ_WIDGET_FRAME, MainFrame,
		EZ_X,(TILES_X * TILE_SIZE) + (FRAME_SIZE * 2),
		EZ_Y, 0, 	
		EZ_WIDTH, STATUS_WIDTH,
		EZ_HEIGHT,MAIN_HEIGHT ,
		EZ_BORDER_WIDTH, FRAME_SIZE,
		EZ_BORDER_TYPE, EZ_BORDER_RAISED,
		EZ_PADX, 0,
		EZ_PADY, 0,
		EZ_ORIENTATION, EZ_VERTICAL_TOP,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		0);													 			
	
	Playfield = EZ_CreateWidget(EZ_WIDGET_LABEL, LeftFrame,
		EZ_X, FRAME_SIZE,
		EZ_Y, FRAME_SIZE,	
		EZ_WIDTH, TILES_X*TILE_SIZE,
		EZ_HEIGHT,TILES_Y*TILE_SIZE,
		EZ_LABEL_PIXMAP, Logo,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,		
		0);			 
// Create status widgets
	
	SubFrame= EZ_CreateWidget(EZ_WIDGET_FRAME, RightFrame,
		EZ_Y,2,
		EZ_ORIENTATION,EZ_VERTICAL_TOP,
		EZ_BORDER_TYPE,EZ_BORDER_RAISED ,
		EZ_BORDER_WIDTH,2,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		0);
	Player = EZ_CreateWidget(EZ_WIDGET_LABEL, SubFrame,
		EZ_LABEL_STRING, PlayerInfo,
		EZ_LABEL_JUSTIFICATION, EZ_LEFT,
		EZ_FONT_NAME,"-Adobe-Helvetica-Bold-R-Normal--*-140-*-*-*-*-*-*",
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		0);																						 			
	Level = EZ_CreateWidget(EZ_WIDGET_LABEL, SubFrame,
		EZ_LABEL_STRING, "Level: ",
		EZ_LABEL_JUSTIFICATION, EZ_LEFT,
		EZ_FONT_NAME,"-Adobe-Helvetica-Bold-R-Normal--*-140-*-*-*-*-*-*",
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		0);													 	
	Score = EZ_CreateWidget(EZ_WIDGET_LCD, SubFrame,
		EZ_WIDTH,128,
		EZ_LCD_NDIGITS, 6,
		EZ_LCD_FONT_SIZE, 16,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BUBBLE_HELP,"Here is your score",
		0);
	EZ_CreateWidget(EZ_WIDGET_LABEL, RightFrame,	
		EZ_HEIGHT, 80,
		0);																																											
	BStartGame = EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, RightFrame,
		EZ_WIDTH, 128,	
		EZ_LABEL_STRING, "Start Game",
		EZ_CALLBACK, CallBackStartGame, NULL,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BORDER_WIDTH, 1,
		EZ_UNDERLINE,6,
		0);
	BUndoMove =	EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, RightFrame,
		EZ_WIDTH, 128,	
		EZ_LABEL_STRING, "Undo Last Move",
		EZ_CALLBACK, CallBackUndoMove, NULL,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BORDER_WIDTH, 1,
		EZ_UNDERLINE,0,
		0);
	EZ_DisableWidget (BUndoMove);															 													 
	BStartLevel = EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, RightFrame,
		EZ_WIDTH, 128,		
		EZ_LABEL_STRING, "Restart Level",
		EZ_UNDERLINE,0,
		EZ_CALLBACK, CallBackRestartLevel,NULL,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BORDER_WIDTH, 1,
		EZ_UNDERLINE,0,
		0);	
	EZ_DisableWidget(BStartLevel);														 													 
	
	EZ_CreateWidget(EZ_WIDGET_LABEL, RightFrame,	
		EZ_HEIGHT, 10,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		0);
	BHelp = EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, RightFrame,
		EZ_WIDTH, 128,		
		EZ_LABEL_STRING, "Help",
		EZ_CALLBACK,CallBackHelp,NULL,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BORDER_WIDTH, 1,
		EZ_UNDERLINE,0,
		0);
	AboutLabel = EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, RightFrame,															
		EZ_WIDTH, 128,
		EZ_LABEL_STRING, "About XSoko",
		EZ_CALLBACK, CallBackAbout, NULL,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BORDER_WIDTH, 1,
		EZ_UNDERLINE,0,
		0);													
	EZ_CreateWidget(EZ_WIDGET_LABEL, RightFrame,	
		EZ_HEIGHT, 10,
		0);												
	BSaveQuit = EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, RightFrame,
		EZ_WIDTH, 128,		
		EZ_LABEL_STRING, "Save",
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BORDER_WIDTH, 1,
		EZ_CALLBACK,CallBackSaveQuit,NULL,
		EZ_UNDERLINE,0,
		0);
	EZ_DisableWidget(BSaveQuit);		
	
	BQuit = EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, RightFrame,
		EZ_WIDTH, 128,		
		EZ_LABEL_STRING, "Quit",
		EZ_CALLBACK, CallBackQuit, NULL,
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BORDER_WIDTH, 1,
		EZ_UNDERLINE,0,
		0);
	
	
	EZ_SetClientMessageHandler(MainEventHandler, NULL);
	EZ_SetApplicationName("XSoko");
	
	InitializeGame();
	
	EZ_DisplayWidget(MainFrame);
	
	LoadGame();
	
	EZ_EventMainLoop();
}
/***********************************************************************************/

// SokoExit()
void SokoExit()
{
	if(GameOn)
	{
		EZ_HideWidget(Playfield);
	}
	EZ_DestroyWidget(MainFrame);
	EZ_Shutdown();
	printf("Quiting XSoko - Have nice day ...\n");
	exit(0);
}



// Initialize Playfield Table

void	InitializePlayfield()
{
	int i,j;
	
	for(i=0;i< TILES_Y;i++)
	{
		for(j=0;j< TILES_X;j++)
		{
			PTable[j][i].typef= PTable[j][i].typeb= '.';
			PTable[j][i].foreground= PTable[j][i].background= NULL;		 	
		}
	}																
	return;	
}


// Load level data 

int	LoadLevelData(char *filename)
{	
	int i,j;
	FILE	*fp;
	char	line[256];
	
	if((fp= fopen(filename,"r"))==NULL)
	{
		return SOKOERROR_OPEN_FILE;
	}
	
	for(i=0;i< TILES_Y;i++)
	{
		if((fgets(line,255,fp)) == NULL)
		{
			fclose(fp);
			
			if(ferror(fp))			
				return SOKOERROR_READ_FILE;			
			else			
				return SOKOERROR_BAD_FORMAT;
		}
		if(strlen(line) < (TILES_X+1))
		{
			fclose(fp);
			return SOKOERROR_BAD_FORMAT;
		}
		for(j=0;j< TILES_X;j++)
		{
			if(line[j]=='o' || line[j]=='+' || line[j]=='x' )
				PTable[j][i].typef=line[j];
			else
				PTable[j][i].typeb=line[j];	 
		}
	}	 
	
	fclose(fp);
	
	return 0;
}

int CreatePlayfielsWidgets()
{
	int i,j;
	int cdest,cbox,chero;
	
	cdest=cbox=chero=0;
	
	for(i=0;i< TILES_Y;i++)
	{
		for(j=0;j< TILES_X;j++)
		{	
		// Create background labels		
			switch(PTable[j][i].typeb)
			{
				case	'.':
					break;				
				case	'=':
					PTable[j][i].background = EZ_CreateWidget(EZ_WIDGET_FREE_LABEL, Playfield,
						EZ_X,j*TILE_SIZE,
						EZ_Y,i*TILE_SIZE,
						EZ_WIDTH,TILE_SIZE,
						EZ_HEIGHT,TILE_SIZE,
						EZ_TRANSIENT, 1,
						EZ_SHAPED_WINDOW, 1,
						EZ_PROPAGATE,0,
						EZ_FREELABEL_MOVABLE,0,
						EZ_LABEL_PIXMAP, Platform,
						EZ_EVENT_HANDLER,GameEventHandler,NULL,		
						0);
					
					cdest++;
					break;
				default:				
					return	SOKOERROR_BAD_FORMAT;
			}	
		// Create foreground (wall) label 	
			switch(PTable[j][i].typef)
			{
				case	'.':
					break;	
				case	'x':
					PTable[j][i].foreground = EZ_CreateWidget(EZ_WIDGET_FREE_LABEL, Playfield,
						EZ_X,j*TILE_SIZE,
						EZ_Y,i*TILE_SIZE,
						EZ_WIDTH,TILE_SIZE,
						EZ_HEIGHT,TILE_SIZE,
						EZ_PROPAGATE,0,																   //EZ_TRANSIENT, 1,
																							 //EZ_SHAPED_WINDOW, 1,
						EZ_FREELABEL_MOVABLE,0,
						EZ_LABEL_PIXMAP, Wall,
						EZ_EVENT_HANDLER,GameEventHandler,NULL,		
						0);
					
					break;							
				case	'o':
					cbox++;
					break;			
				case	'+':					
					chero++;
					HeroPosX=j;HeroPosY=i;
					break;
				default:
					return	SOKOERROR_BAD_FORMAT;
			}
		}
	}
	if((chero!= 1) || (cbox!= cdest) || (cbox==0))
	{
		return SOKOERROR_BAD_FORMAT;
	}
// Now create block labels  	
	Barrels=cbox;
	for(i=0;i< TILES_Y;i++)
	{
		for(j=0;j< TILES_X;j++)
		{	
			if(PTable[j][i].typef== 'o')
			{
				PTable[j][i].foreground = EZ_CreateWidget(EZ_WIDGET_FREE_LABEL, Playfield,
					EZ_X,j*TILE_SIZE,
					EZ_Y,i*TILE_SIZE,
					EZ_WIDTH,TILE_SIZE,
					EZ_HEIGHT,TILE_SIZE,
					EZ_TRANSIENT, 1,
					EZ_SHAPED_WINDOW, 1,
					EZ_FREELABEL_MOVABLE,0,
					EZ_LABEL_PIXMAP, Object,
					EZ_EVENT_HANDLER,GameEventHandler,NULL,
					EZ_BUBBLE_HELP,"Push me !",
					EZ_PROPAGATE,0,
					0);
				
				
			}																				 
		}
	}
	
// And lastly create hero label 
	
	PTable[HeroPosX][HeroPosY].foreground = EZ_CreateWidget(EZ_WIDGET_FREE_LABEL, Playfield,
		EZ_X,HeroPosX*TILE_SIZE,
		EZ_Y,HeroPosY*TILE_SIZE,
		EZ_WIDTH,TILE_SIZE,
		EZ_HEIGHT,TILE_SIZE,
		EZ_PADX,0,
		EZ_PADY,0,
		EZ_TRANSIENT, 1,
		EZ_SHAPED_WINDOW, 1,
		EZ_FREELABEL_MOVABLE,0,
		EZ_PROPAGATE,0,
		EZ_LABEL_PIXMAP, Hero[0],
		EZ_EVENT_HANDLER,GameEventHandler,NULL,
		EZ_BUBBLE_HELP,"It is you.",		
		0);
	
	EZ_DisplayWidget(Playfield);
	return 0;
}

void DeletePlayfieldWidgets()
{
	int i,j;
	
	EZ_HideWidget(Playfield);
	for(i=0;i< TILES_Y;i++)
	{
		for(j=0;j< TILES_X;j++)
		{
			if(PTable[j][i].foreground != NULL)
			{
				
				EZ_DestroyWidget(PTable[j][i].foreground);
			}
			if(PTable[j][i].background != NULL)
			{
				
				EZ_DestroyWidget(PTable[j][i].background);		
			}
		}
	}
	
	EZ_DisplayWidget(Playfield);
	return;
}


static	void MainEventHandler(EZ_Widget *widget, void *data, int etype, XEvent *xeve)
{
	
	XClientMessageEvent *ev = (XClientMessageEvent *)xeve;      
	
	if(ev->message_type == WMProtocolsAtom)
	{
		Atom c = (ev->data.l)[0];
		if(c == DeleteWindowAtom)
		{
			if(widget!= MainFrame)
				return;
			if(!GameOn)
				SokoExit();	
			if(DisplayDialog("Are you sure you want to quit?","Yep","No",NULL,"XSoko Confirmation"))
				SokoExit();						
			return;
		}
	}	
}	
static	void GameEventHandler(EZ_Widget *widget, void *data, int etype, XEvent *xeve)
{
	int tmpposX,tmpposY,result,move,direction;
	int offx2,offy2;
	
	if(!GameOn)
		return;
	
	offx2=offy2=0;	
	tmpposX=HeroPosX; tmpposY=HeroPosY;
	
	switch(etype)
	{		
		case	EZ_KEY_PRESS:
			
			move=0;	
			switch(EZ_PressedKey)
			{
				case	EZ_LEFT_KEY:
					direction= SOKO_LEFT;													// left		
					tmpposX--;offx2--;
					break;
				case	EZ_RIGHT_KEY:
					direction= SOKO_RIGHT;				
					tmpposX++;offx2++;
					break;
				case  EZ_UP_KEY:
					direction= SOKO_UP;				
					tmpposY--;offy2--;
					break;
				case	EZ_DOWN_KEY:
					direction= SOKO_DOWN;			
					tmpposY++;offy2++;
					break;
				case	'u':
				case  'U':
					if(LOldHX!=255)
					{
						RedrawBlockPosition(HeroPosX, HeroPosY, LOldHX, LOldHY);
						HeroPosX= LOldHX; HeroPosY= LOldHY;
						LOldHX=255;
						EZ_DisableWidget (BUndoMove);	
					}
					if(LOldBX!=255)
					{
						RedrawBlockPosition(LNewBX, LNewBY, LOldBX, LOldBY);
						LOldBX=LOldBY=255;	
					}
					return;
				default:				
					return;		
			}		
			EZ_ConfigureWidget(PTable[HeroPosX][HeroPosY].foreground,EZ_LABEL_PIXMAP,Hero[direction],0);	
			result = ValidateHeroPosition(tmpposX,tmpposY,direction);		
			if(result == 3)
				return;
			LOldBX=LOldBY=255;			
			if(result == 1)
			{
				RedrawBlockPosition(tmpposX, tmpposY, tmpposX+offx2, tmpposY+offy2);
				LOldBX=tmpposX; LOldBY=tmpposY;
				LNewBX=tmpposX+offx2; LNewBY=tmpposY+offy2;
			}	
			Moves++;				
			RedrawBlockPosition(tmpposX-offx2, tmpposY-offy2, tmpposX, tmpposY);
			LOldHX=tmpposX-offx2; LOldHY=tmpposY-offy2;	
			HeroPosX=tmpposX; HeroPosY=tmpposY;
			EZ_EnableWidget (BUndoMove);						
			
			if(result==1)
			{
				if(IsLevelDone())
				{
					int bonus=(550+(5*GameLevel)+(80*Barrels));
					char	info[128];
					EZ_SitForALittleBit(20000);
					if(Moves<bonus)
					{
						char	tmp[6];
						Scorew= Scorew+(bonus-Moves);
						sprintf(tmp,"%06d",Scorew);
						EZ_SetLcdString(Score,tmp);
					}
					sprintf(info,"Level %d finished after %d moves.\n",GameLevel,Moves);
					DisplayDialog(info , "OK", NULL,NULL,"XSoko -information");
					NextLevel();
				}	
			}
			break;
			
		default:
			break;							
	}
	return;
}

// Validate new hero position 
// returns:
//  				0 = valid move
//  				1 = valid move that will generate push 
//  				3 = invalid move

int ValidateHeroPosition(int posX,int posY,int direction)
{
	
	if(PTable[posX][posY].typef == '.') 	// check for empty space		
		return  0;
	if((posX<0 || posX == TILES_X) || (posY<0 || posY == TILES_Y))	// check for boundaries
		return	3;		
	if(PTable[posX][posY].typef == 'o') // if block then see if it can be moved	
	{
		switch(direction)
		{
			case	SOKO_LEFT:
				return ValidateBlockPosition(posX-1,posY);
			case  SOKO_RIGHT:
				return ValidateBlockPosition(posX+1,posY);
			case  SOKO_UP:
				return ValidateBlockPosition(posX,posY-1);		 					
			case  SOKO_DOWN:
				return ValidateBlockPosition(posX,posY+1);
		}
	}
	return	3;	
}

int ValidateBlockPosition(int posX,int posY)
{
	
	if((posX<0 || posX == TILES_X) || (posY<0 || posY == TILES_Y))
		return	3;
	if(PTable[posX][posY].typef == '.')
		return  1;
	return	3;		
}

void	RedrawBlockPosition(int oldx, int oldy, int newx, int newy)
{
	
	PTable[newx][newy].foreground = PTable[oldx][oldy].foreground;
	PTable[newx][newy].typef = PTable[oldx][oldy].typef;	
	
	PTable[oldx][oldy].foreground= NULL;
	PTable[oldx][oldy].typef= '.';
	EZ_ConfigureWidget(PTable[newx][newy].foreground, 
		EZ_X, newx* TILE_SIZE,
		EZ_Y, newy* TILE_SIZE,
		0);								 
	return;
}


int	IsLevelDone()
{
	int i,j;
	
	for(i=0;i< TILES_Y;i++)
	{
		for(j=0;j< TILES_X;j++)
		{
			if(PTable[j][i].typef == 'o' && PTable[j][i].typeb != '=')
				return 0;		
		}		
	}
	return 1;
}	

void	StopGame()
{
	if(!GameOn)
		return;	
	GameOn= 0;
	EZ_HideWidget(Playfield);
	DeletePlayfieldWidgets();
	DisplayPlayground(0);	
	InitializeGame();
	EZ_ConfigureWidget(BStartGame,EZ_LABEL_STRING,"Start Game",EZ_UNDERLINE,6,0);
	EZ_DisableWidget (BStartLevel);
	EZ_DisableWidget (BSaveQuit);
	EZ_DisableWidget (BUndoMove);
	return;
}

void	StartGame()
{
	if(GameOn)
		return;
	GameOn= 1;
	
	InitializeGame();
	DisplayPlayground(1);
	EZ_ConfigureWidget(BStartGame,EZ_LABEL_STRING,"End Game",EZ_UNDERLINE,0,0);
	EZ_EnableWidget (BStartLevel);
	EZ_EnableWidget (BSaveQuit);
	EZ_SetFocusTo(Playfield);
	NextLevel();	
	return;
}	

void	RestartLevel()
{
	if(!GameOn)
		return;
	if(Scorew)
	{
		char	tmp[6];
		if(Scorew>49)
			Scorew= Scorew-50;	
		else
			Scorew=0;
		sprintf(tmp,"%d",Scorew);
		EZ_SetLcdString(Score,tmp);
	}	
	GameLevel--;
	NextLevel();
	return;	
}		

void	CallBackStartGame(EZ_Widget *widget, void *data)
{
	if(GameOn)
	{	
		if(!(DisplayDialog("Are you sure you want to end this game ?","Yes","No",NULL,"XSoko Confirmation")))
			return;
		StopGame();		
	}			
	else
	{
		StartGame();
	}			
	return;	
}		

void	CallBackUndoMove(EZ_Widget *widget, void *data)
{
	if(LOldHX!=255)
	{
		RedrawBlockPosition(HeroPosX, HeroPosY, LOldHX, LOldHY);
		HeroPosX= LOldHX; HeroPosY= LOldHY;
		LOldHX=255;
	}
	if(LOldBX!=255)
	{
		RedrawBlockPosition(LNewBX, LNewBY, LOldBX, LOldBY);
		LOldBX=LOldBY=255;	
	}
	EZ_DisableWidget (BUndoMove);	
	return;	
}

void  CallBackRestartLevel(EZ_Widget *widget, void *data)
{
	
	if(!(DisplayDialog("Are you sure you want to restart this level ?","Yes","No",NULL,"XSoko Confirmation")))
		return;
	
	RestartLevel();	
	return;	
}

void  CallBackSaveQuit(EZ_Widget *widget, void *data)
{
	if(SaveCurrentGame()!=0)
	{
		DisplayDialog("Unable to save game.","OK",NULL,NULL,"XSoko Error");
	}
	return;
}


void	CallBackQuit(EZ_Widget *widget, void *data)
{
	if(GameOn)
	{
		if(DisplayDialog("Are you sure you want to quit?","Yep","No",NULL,"XSoko Confirmation"))
			SokoExit();
		else
			return;
	}
	SokoExit();
}

void	CallBackAbout(EZ_Widget *widget, void *data)
{
	char	tmp[255];
	
	strcpy(tmp,"- XSoko -\n by W. Rawdanik\n");
	strcat(tmp,"-------------------\nThis program was written using M.Zou's EZ Widget and Graphics Library (EZWGL)\n");
	strcat(tmp,"\n(www.ma.utexas/~mzou/EZWGL)");
	
	DisplayDialog(tmp,"OK",NULL,NULL,"XSoko About");					  
	return;
}

void	CallBackHelp(EZ_Widget *widget, void *data)
{
	DisplayHelpWindow();
	return;
}

void	DisplayPlayground(int mode)
{
	if(mode)	
	{
		
		EZ_ConfigureWidget(Playfield,	EZ_LABEL_PIXMAP, Background,0);
		return;
	}	
	else
	{
		EZ_ConfigureWidget(Playfield,	EZ_LABEL_PIXMAP, Logo,0);		
		return;
	}
}


int	DisplayDialog(char *info, char *buttonA, char *buttonB, char *buttonC,char *title)
{	
	EZ_Widget *Dialog, *DialogFrame,*DialogLabel,*BA, *BB,*BC;
	static int answer;
	
	answer = 255;	
	Dialog= EZ_CreateWidget(EZ_WIDGET_FRAME, NULL,
		EZ_TRANSIENT, 1,
		EZ_WMHINTS,PMinSize|PMaxSize,
		EZ_IPADX,4,
		EZ_IPADY,0,
										  // 	 EZ_ORIENTATION,EZ_VERTICAL_TOP,
		0);
	DialogFrame = EZ_CreateWidget(EZ_WIDGET_FRAME,  Dialog,
		EZ_BORDER_TYPE,EZ_BORDER_SUNKEN,
		EZ_BORDER_WIDTH,1, 
		0);															
	DialogLabel = EZ_CreateWidget(EZ_WIDGET_LABEL,DialogFrame,
		EZ_LABEL_JUSTIFICATION, EZ_CENTER,		
		0);																																											
	BA= EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, Dialog,
		EZ_LABEL_STRING, buttonA,
		EZ_CALLBACK, DialogCallbackA, &answer,
		EZ_BORDER_WIDTH, 1,
		EZ_UNDERLINE,0,
		0);		
	if(buttonB)
	{																							
		BB=EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, Dialog,
			EZ_LABEL_STRING, buttonB,
			EZ_CALLBACK, DialogCallbackB, &answer,
			EZ_BORDER_WIDTH, 1,
			EZ_UNDERLINE,0,
			0);																	
	}		
	if(buttonC)
	{																							
		BC=EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, Dialog,
			EZ_LABEL_STRING, buttonC,
			EZ_CALLBACK, DialogCallbackC, &answer,
			EZ_BORDER_WIDTH, 1,
			EZ_UNDERLINE,0,
			0);																	
	}	
	
	EZ_ConfigureWidget(DialogLabel,EZ_LABEL_STRING,info,0);
	EZ_DisplayWidget(Dialog);
	
	if(title)
		XStoreName((Display *)EZ_GetDisplay(),EZ_GetWidgetWindow(Dialog),title);
	
	EZ_SetGrab(Dialog);
	
	if(buttonB)
		EZ_SetFocusTo(BB);
	else
		EZ_SetFocusTo(BA);
	
	while(answer==255)
	{
		EZ_WaitAndServiceNextEvent();
	}
	EZ_ReleaseGrab();
	EZ_DestroyWidget(Dialog);
	
	return answer;
}

void	DialogCallbackA(EZ_Widget *widget, int	*answer)
{
	*answer = 1;
	return;
}
void	DialogCallbackB(EZ_Widget *widget, int	*answer)
{
	*answer = 0;
	return;
}
void	DialogCallbackC(EZ_Widget *widget, int	*answer)
{
	*answer = 2;
	return;
}

void NextLevel()
{
	char	filename[256];
	char	line[4];
	FILE	*fp;
	
	GameLevel++;
	Moves=0;
	LOldBX=LOldBY=255;
	LOldHX=LOldHY=255;
	if(GameLevel>9)
		sprintf(filename,"%s/%slevel_%d.txt",path,SOKO_MAPS,GameLevel);
	else
		sprintf(filename,"%s/%slevel_0%d.txt",path,SOKO_MAPS,GameLevel);
	if((fp= fopen(filename,"r"))==NULL)
	{
		DisplayDialog("Unable to find next level file !","OK",NULL,NULL,"XSoko Error");
		StopGame();
		return;
	}
	
	if((fgets(line,4,fp)) == NULL)
	{
		fclose(fp);
		DisplayDialog("Unable to load next level file !","OK",NULL,NULL,"XSoko Error");
		StopGame();
		return;
	}
	*(line+3)=0;
	
	fclose(fp);
	if(strcmp(line,"END"))	
	{
		char	tmp[4];
		
		DeletePlayfieldWidgets();
		InitializePlayfield();
		
		if(LoadLevelData(filename) !=0)
		{	
			DisplayDialog("Unable to load level data !","OK",NULL,NULL,"XSoko Error");	
			StopGame();		
			return;
		}	
		
		if((CreatePlayfielsWidgets()) !=0)
		{				
			DisplayDialog("Corrupted level data !","OK",NULL,NULL,"XSoko Error");
			StopGame();				
			return;		
		}
		
		sprintf(tmp,"Level: %d",GameLevel);
		EZ_ConfigureWidget(Level, EZ_LABEL_STRING, tmp,0);
		EZ_DisableWidget (BUndoMove);				
		return;
	}	
	else
	{	
		char	tmp[128];
		sprintf(tmp,"Well, that was the last level.\nYour Score is %d\nCONGRATULATION !!!",Scorew);	
		DisplayDialog(tmp,"OK",NULL,NULL,"XSoko Game Over");
		StopGame();					
		return;
	}
}	

void	InitializeGame()
{
	char	tmp[10];
	
	LOldBX=LOldBY=255;
	LOldHX=LOldHY=255;
	GameLevel=Scorew=0;
	EZ_SetLcdString(Score,"000000");
	EZ_ConfigureWidget(Level, EZ_LABEL_STRING, "Level: 0",0);
	return;
}

int CreatePixmaps()
{
	int	result=0;	
	char	tmp[512];
	sprintf(tmp,"%s/%s/block.gif",path,SOKO_GRAPH);
	if((Wall=EZ_CreateLabelPixmapFromImageFile (tmp))== NULL)
		result++;
	sprintf(tmp,"%s/%s/logo.gif",path,SOKO_GRAPH);
	if((Logo=EZ_CreateLabelPixmapFromImageFile (tmp))== NULL)
		result++;
	sprintf(tmp,"%s/%s/lin64.gif",path,SOKO_GRAPH);
	if((Background=EZ_CreateLabelPixmapFromImageFile (tmp))== NULL)
		result++;
	sprintf(tmp,"%s/%s/podstawka.gif",path,SOKO_GRAPH);
	if((Platform=EZ_CreateLabelPixmapFromImageFile (tmp))== NULL)
		result++;
	sprintf(tmp,"%s/%s/beczka.xpm",path,SOKO_GRAPH);
	if((Object=EZ_CreateLabelPixmapFromXpmFile (tmp))== NULL)
		result++;		
	
	// for now use one frame for all directions ... 
	
	sprintf(tmp,"%s/%s/pingwin.xpm",path,SOKO_GRAPH);
	if((Hero[3]=EZ_CreateLabelPixmapFromXpmFile (tmp))== NULL)
		result++;
	if((Hero[2]=EZ_CreateLabelPixmapFromXpmFile (tmp))== NULL)
		result++;
	if((Hero[0]=EZ_CreateLabelPixmapFromXpmFile (tmp))== NULL)
		result++;
	if((Hero[1]=EZ_CreateLabelPixmapFromXpmFile (tmp))== NULL)
		result++;	
	
	
	return result;	
}

void	DisplayHelpWindow()
{
	EZ_Widget *TextFrame;
	EZ_Widget *BDone;
	EZ_Widget *TextWindow;
	char	tmp[512];
	
	TextFrame = EZ_CreateWidget(EZ_WIDGET_FRAME, NULL,
		EZ_WIDTH,540,
		EZ_HEIGHT,400,
		EZ_ORIENTATION,EZ_VERTICAL_TOP,
		EZ_FILL_MODE,EZ_FILL_BOTH,
		EZ_TRANSIENT,1, 	
		0);
	TextWindow= EZ_CreateTextWidget(TextFrame, 0,1,1);
	EZ_ConfigureWidget(TextWindow,
		EZ_LABEL_PIXMAP, Background,
		EZ_FOREGROUND,"white",
		EZ_SELECTION_BACKGROUND, "black",
		EZ_BORDER_WIDTH, 1,
		0); 
	BDone = EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, TextFrame,
		EZ_HEIGHT,32,
		EZ_LABEL_STRING, "Close Help Window",
		EZ_CALLBACK, HelpCallbackCloseWindow,TextFrame,
		EZ_BORDER_WIDTH, 1,
		0);
	
	sprintf(tmp,"%s/help.txt",path);
	EZ_TextLoadAnnotatedFile(TextWindow,tmp);
	EZ_DisplayWidget(TextFrame);
	EZ_SetFocusTo(BDone); 
	XStoreName((Display *)EZ_GetDisplay(),EZ_GetWidgetWindow(TextFrame),"XSoko - Help");
	return;
}	

void	HelpCallbackCloseWindow(EZ_Widget *widget, void *data)
{
	EZ_DestroyWidget((EZ_Widget *) data);
	return;
}

int SaveCurrentGame()
{
	FILE	*fp;
	char	filename[256];
	int 	i,j;
	
	
	strcpy(filename, PlayerHome);
	strcat(filename,"/.XSokoLevel");	
	if((fp= fopen(filename,"w"))==NULL)
	{
		return SOKOERROR_OPEN_FILE;
	}
	
	fprintf(fp, "%d,%d,%d",GameLevel,Scorew,Moves);
	fclose(fp);
	
	strcpy(filename, PlayerHome);
	strcat(filename,"/.XSokoData");	
	if((fp= fopen(filename,"w"))==NULL)
	{
		return SOKOERROR_OPEN_FILE;
	}
	
	for(i=0;i< TILES_Y;i++)
	{
		for(j=0;j< TILES_X;j++)
		{	
			fputc(PTable[j][i].typef,fp);
			fputc(PTable[j][i].typeb,fp);
		}
	}
	
	fclose(fp);
	return	0;	
}

int LoadGame()
{
	FILE	*fp;
	char	filename[256];
	char	tmp[128];
	int 	i,j;
	
	strcpy(filename, PlayerHome);
	strcat(filename,"/.XSokoLevel");	
	if((fp= fopen(filename,"r"))==NULL)
	{
		return SOKO_NO_SAVED_GAME;
	}
	
	fscanf(fp,"%d,%d,%d",&GameLevel,&Scorew,&Moves);
	fclose(fp);
	
	sprintf(tmp,"There is a saved game in your home directory.\nLevel=%d, Score=%d\n",GameLevel,Scorew);
	strcat(tmp,"Do you want to load this game?");
	i= DisplayDialog(tmp,"Yes","No","Delete Saved Game","XSoko Confirmation");
	if(i==0)
		return;
	else
	{
		if(i==2)
		{
			unlink(filename);
			strcpy(filename, PlayerHome);
			strcat(filename,"/.XSokoData");
			unlink(filename);
			return;	
		}
	}
	
	strcpy(filename, PlayerHome);
	strcat(filename,"/.XSokoData");	
	if((fp= fopen(filename,"r"))==NULL)
	{
		DisplayDialog("Unable to open save file !","OK",NULL,NULL,"XSoko Error");
		InitializeGame();
		return SOKOERROR_OPEN_FILE;
	}
	
	for(i=0;i< TILES_Y;i++)
	{
		for(j=0;j< TILES_X;j++)
		{				
			PTable[j][i].typef=fgetc(fp);
			PTable[j][i].typeb=fgetc(fp);
			if(PTable[j][i].typef=='o')
				Barrels++;
		}
	}
	fclose(fp);
	
	DisplayPlayground(1);
	if((CreatePlayfielsWidgets()) !=0)
	{				
		DisplayDialog("Corrupted level data !","OK",NULL,NULL,"XSoko Error");
		StopGame();				
		return;		
	}
	GameOn= 1;
	LOldBX=LOldBY=255;
	LOldHX=LOldHY=255;
	
	EZ_ConfigureWidget(BStartGame,EZ_LABEL_STRING,"End Game",0);
	EZ_EnableWidget (BStartLevel);
	EZ_EnableWidget (BSaveQuit);
	EZ_DisableWidget (BUndoMove);	
	EZ_SetFocusTo(Playfield);
	sprintf(tmp,"%06d",Scorew);
	EZ_SetLcdString(Score,tmp);
	sprintf(tmp,"Level: %d",GameLevel);
	EZ_ConfigureWidget(Level, EZ_LABEL_STRING, tmp,0);		
	return	0;
}

